/* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 932 $ $LastChangedDate: 2002-09-11 17:06:10 +0200 (Wed, 11 Sep 2002) $
 *
 * Runtime environment for gont.
 */

#include <gont/runtime.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define ALLOCATOR_STATS 1
/* Define for example if you want to play with efence... */
/* #define GONT_USE_MALLOC */

#ifdef GONT_USE_MALLOC
#  define GC_malloc malloc
#  define GC_malloc_atomic malloc
#  define GC_register_finalizer(o, fn, cd, ofn, ocd)  do {} while (0)
#  undef GONT_USE_GC_MALLOC_MANY
#  undef GONT_USE_GC
#else
#  ifdef GONT_USE_GC
#    include <gc/gc.h>
#  else
#    error "Neither GONT_USE_GC nor GONT_USE_MALLOC defined."
#  endif
#endif

struct stats {
	gont_int normal_cnt;
	gont_int normal_bytes;
	gont_int atomic_cnt;
	gont_int atomic_bytes;
	gont_int special_cnt;
	gont_int special_bytes;
	gont_int max_heap_size;
};

static struct stats allocator_stats;

static void oops(int s) __attribute__ ((__noreturn__));

static void oops(int s)
{
	gont_oops("Run out of memory, trying to allocate %d bytes.\n", s);
}

#ifdef GONT_USE_GC
#  if ALLOCATOR_STATS
#    define RECORD_HEAP_SIZE() \
	do { \
		if (GC_get_heap_size() > allocator_stats.max_heap_size) \
			allocator_stats.max_heap_size = GC_get_heap_size(); \
	} while (0)
#  endif
#endif

#ifndef RECORD_HEAP_SIZE
#  define RECORD_HEAP_SIZE() do { } while (0)
#endif

#ifdef GONT_USE_GC_MALLOC_MANY
/* The values 3 and 2 were tuned for duron running linux, and Gont compiler
 * as tested application.  You might experiment with different values, but
 * remember not to set MAX_SPECIAL_BYTES_ATOMIC above MAX_SPECIAL_BYTES.
 */
#define WORD_SIZE (sizeof(gont_int))
#define MAX_SPECIAL_WORDS 3
#define MAX_SPECIAL_BYTES (MAX_SPECIAL_WORDS * WORD_SIZE)
#define MAX_SPECIAL_BYTES_ATOMIC (2 * WORD_SIZE)

static void *special_sizes[MAX_SPECIAL_WORDS];

static void *alloc_special(gont_int sz)
{
	gont_int aligned = (sz + (WORD_SIZE - 1)) & ~(WORD_SIZE - 1);
	int off = (aligned / WORD_SIZE) - 1;
	void **p;
	extern GC_PTR GC_malloc_many(size_t);

	p = special_sizes[off];

	if (p == NULL) {
		p = GC_malloc_many(aligned);
		RECORD_HEAP_SIZE();
		if (p == NULL)
			oops(sz);
	}

	special_sizes[off] = *p;
	*p = NULL;

#if ALLOCATOR_STATS
	allocator_stats.special_cnt++;
	allocator_stats.special_bytes += aligned;
#endif /* ALLOCATOR_STATS */

	return p;
}
#endif /* GONT_USE_GC_MALLOC_MANY */

void gont_alloc_dump_stats(void)
{
#define M (1024.0 * 1024.0)
	printf("allocator statistics:\n");
#if ALLOCATOR_STATS
	printf("   normal: %10ld allocations, %.2f Mb\n"
	       "   atomic: %10ld allocations, %.2f Mb\n"
	       "  special: %10ld allocations, %.2f Mb\n"
	       "  maximal GC heap size: %.3f Mb\n",
	       allocator_stats.normal_cnt,
	       allocator_stats.normal_bytes / M,
	       allocator_stats.atomic_cnt,
	       allocator_stats.atomic_bytes / M,
	       allocator_stats.special_cnt,
	       allocator_stats.special_bytes / M,
	       allocator_stats.max_heap_size / M);
#else
	printf("  ALLOCATOR_STATS set to 0.\n");
#endif
#ifdef GONT_USE_GC
	RECORD_HEAP_SIZE();
	printf("  heap size: %.3f Mb, free: %.3f Mb, total: %.3f Mb\n",
	       GC_get_heap_size() / M, 
	       GC_get_free_bytes() / M,
	       GC_get_total_bytes() / M);
#endif
#undef M
}

void *gont_xmalloc(gont_int s)
{
	void *p;

#ifdef GONT_USE_GC_MALLOC_MANY
	if (s <= MAX_SPECIAL_BYTES)
		return alloc_special(s);
#endif /* GONT_USE_GC_MALLOC_MANY */

#if ALLOCATOR_STATS
	allocator_stats.normal_cnt++;
	allocator_stats.normal_bytes += s;
#endif /* ALLOCATOR_STATS */

	p = GC_malloc(s);
	RECORD_HEAP_SIZE();

	if (p == NULL) 
		oops(s);
	else
		return p;
}

void *gont_xmalloc0(gont_int s)
{
	return memset(gont_xmalloc(s), '\0', s);
}

void *gont_xmalloc_atomic(gont_int s)
{
	void *p;

#ifdef GONT_USE_GC_MALLOC_MANY
	if (s <= MAX_SPECIAL_BYTES_ATOMIC)
		return alloc_special(s);
#endif /* GONT_USE_GC_MALLOC_MANY */

#if ALLOCATOR_STATS
	allocator_stats.atomic_cnt++;
	allocator_stats.atomic_bytes += s;
#endif /* ALLOCATOR_STATS */

	p = GC_malloc(s);
	RECORD_HEAP_SIZE();

	if (p == NULL) 
		oops(s);
	else
		return p;
}

void *gont_xmalloc0_atomic(gont_int s)
{
	return memset(gont_xmalloc_atomic(s), '\0', s);
}

void gont_register_finalizer(void *obj, void (*fin)(void *obj, void *data), 
			     void *data)
{
	GC_register_finalizer(obj, fin, data, NULL, NULL);
}

void gont_init_memory_allocator()
{
}
