/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 920 $ $LastChangedDate: 2002-09-10 10:03:29 +0200 (Tue, 10 Sep 2002) $
 *
 * Bind - Name binder. 
 * 
 * In general it changes all occurrences of Simple_name(_) and 
 * Qualified_name(_) to Symbol(_), so next phases of compilation don't need 
 * to care about symbol tables, scopes and so forth.
 *
 * In addition we do some closure marking stuff here. Variables are
 * marked to be put into closure, closurized variables lists are created
 * for each function (including lambda functions), each reference to
 * symbol annotated with depth of walk in closure stack needed to get
 * to it.
 */

/* someday-TODO: check if it's still true ;)
 * Level stuff:
 *   fnc_level is level of function nesting for symbol.
 *     -1 for symbols from other modules
 *	0 for global symbols (in current module)
 *	1 for variables and parameters to 1st level functions
 *	2 ditto for 2nd level...
 *   scope_level is level of scope nesting of symbol, levels -2 up to
 *   0 are the same as for fnc_level, next define level of block nesting.
 *
 * Example	      // (fnc_level, scope_level)
 * int foo;	      // (0, 0)
 * int bar(	      // (0, 0)
 *     int x) {	      // (1, 1)
 *     int baz;	      // (1, 1)
 *     int f() {      // (1, 1)
 *	   int k;     // (2, 2)
 *     }
 *     { int z; }     // (1, 2)
 *  }
 */


// speed-TODO: maybe do some kind of copy-on-write, instead of always
// copying hash of local symbols, or cache few symbols in list.
// Ok, it's to be investigated.

open Ast;
open Util;
open Tyutil;

struct bind_ctx {
    int current_fnc_level;
    int current_scope_level;
    int current_loop_level;
   
    <var_symbol>Hashtbl::t local_vars;
    <li_symbol>Hashtbl::t loop_ids;
    <li_symbol>option top_loop_id;

    <var_symbol>Hashtbl::t global_vars;
    <ty_symbol>Hashtbl::t tydefs;
    <sf_symbol>Hashtbl::t struct_fields;

    // array refers to function levels
    <<var_symbol>list>Array::t closures;

    <var_symbol>list formal_parms;

    <id>list open_modules;
  
    <void>Hashtbl::t referenced_modules;
    <void>Hashtbl::t used_modules;
    <void>Hashtbl::t used_exns;
    <void>Hashtbl::t used_open_modules;
    
    <id>list bound_tyvars;

    bool needs_closure;

    // variables bound by current pattern
    <var_symbol>list pattern_vars;
    // variables bound in current scope, used for unused-checks
    // this includes variables bound in patterns
    <var_symbol>list scope_vars;

    <bind_ctx>option parent;
}

bool done_init;
bind_ctx global_ctx;

section init
{
    done_init = false;
    global_ctx = {
	current_fnc_level = 0,
	current_scope_level = 0,
	current_loop_level = 0,
    
	local_vars = Hashtbl::create(13),
	loop_ids = Hashtbl::create(13),
	top_loop_id = None,
	open_modules = Nil,

	bound_tyvars = Nil,

	formal_parms = Nil,

	global_vars = Hashtbl::create(13),
	tydefs = Hashtbl::create(13),
	struct_fields = Hashtbl::create(13),

	closures = Array::make(10, Nil),

	used_exns = Hashtbl::create(13),
	used_modules = Hashtbl::create(13),
	referenced_modules = Hashtbl::create(13),
	used_open_modules = Hashtbl::create(13),

	needs_closure = true,

	pattern_vars = Nil,
	scope_vars = Nil,
	
	parent = None
    };
}

// reference_module() is used with Open and when types from module are mentioned
void reference_module(id m)
{
    if (!Hashtbl::mem(global_ctx.referenced_modules, m))
	Hashtbl::add(global_ctx.referenced_modules, m, ());
}

// to use_module() means to need variable or function from it
void use_module(id m)
{
    if (!Hashtbl::mem(global_ctx.used_modules, m)) {
	Hashtbl::add(global_ctx.used_modules, m, ());
	// used => referenced
	reference_module(m);
    }
}

void init_binder(id mod_name, bool iface)
{
    current_module = mod_name;
    doing_iface = iface;

    _ core = "Core";
    if (mod_name != core) {
	global_ctx.open_modules = List::cons1(core);
	Module::open_module(core);
 	reference_module(core);
	// don't warn about unused 'open Core;'
	Hashtbl::add(global_ctx.used_open_modules, core, ());
    }

    done_init = true;
}

void check_init()
{
    if (!done_init)
	ice("Bind::check_init(): didn't Bind::init_binder()!");
}

bind_ctx copy_context(bind_ctx parent)
{
    return {
	current_scope_level = parent.current_scope_level + 1,
	parent = Some(parent),
	loop_ids = Hashtbl::copy(parent.loop_ids),
	local_vars = Hashtbl::copy(parent.local_vars),
	scope_vars = [],
    
	// rest is simply copied:
	current_fnc_level = parent.current_fnc_level,
	current_loop_level = parent.current_loop_level,
	global_vars = parent.global_vars,
	tydefs = parent.tydefs,
	closures = parent.closures,
	struct_fields = parent.struct_fields,
	bound_tyvars = parent.bound_tyvars,
	formal_parms = parent.formal_parms,
	used_exns = parent.used_exns,
	used_modules = parent.used_modules,
	referenced_modules = parent.referenced_modules,
	used_open_modules = parent.used_open_modules,
	open_modules = parent.open_modules,
	pattern_vars = parent.pattern_vars,
	top_loop_id = parent.top_loop_id,
	needs_closure = false
    };
}

void check_if_in_current_module(bind_ctx, <'a>symbol s)
{
    if (s.module != "" && s.module != current_module) {
	error("cannot define symbols from module `" + s.module +
	      "' in `" + current_module + "'");
	s.module = current_module;
    }
}

*(string, string) fetch_module_name(bind_ctx c, <'a>bindable src)
{
    switch src {
    case Simple_name(n):
	// if in global scoping, must prepend current module
	// otherwise it is not necessary
	if (c.current_scope_level == 0) 
	    return (current_module, n);
	else
	    return ("", n);
	
    case Qualified_name(m, n):
	return (m, n);
	
    case Symbol(_):
	return ice("fetch_module_name(): already made");
    }
}

// We need to remove type variables from upper-level function here. 
// I.e. in void f('a) { void g('a, 'b) {}}  for g it ft.parms should 
// be [b], not [a,b] as suggested by parser.
<id>list fixup_tyvars(bind_ctx c, <id>list tyvars)
{
    bool isnt_global_tyvar(id n)
    {
	return !List::exists(fun bool (id x) (x == n), c.bound_tyvars);
    }
    
    return List::filter(isnt_global_tyvar, tyvars);
    
}

// someday-TODO: make warning if new structure extends or narrows
// old one (i.e. there is no way one may make instance of it with
// { ... })
void make_struct_field(bind_ctx c, ty_symbol tydef, <sf_symbol>field fld)
{
    push_location(fld.fld_loc);
    
    try {
	sf_symbol s;
	
	let (module, name) = fetch_module_name(c, fld.fld_symbol) in
	s = {module = module, 
	     name = name, 
	     stype = fld.fld_type,
	     sloc = top_location(),
	     a = {sf_tydef = tydef}};
    
	check_if_in_current_module(c, s);

	Hashtbl::add(c.struct_fields, s.name, s);

	fld.fld_symbol = Symbol(s);
	
    } finally {
	pop_location();
    }
}

var_symbol make_var(bind_ctx c, ty t, <var_symbol>bindable src)
{
    string module, name, xc_name;

    (module, name) = fetch_module_name(c, src);

    if (module == "")
	xc_name = name;
    else
	xc_name = module + "$$" + name;

    var_symbol s = {module = module, 
		    name = name, 
		    stype = t,
		    sloc = top_location(),
		    a = {fnc_level = c.current_fnc_level,
			 scope_level = c.current_scope_level,
			 properties = 0,
			 parent_union = None,
			 extern_name = xc_name}};
    
    check_if_in_current_module(c, s);

    List::iter((fun void (var_symbol x) {
	    if (x.a.scope_level == s.a.scope_level) {
		error("redefinition of symbol `" + name + "'");
	    } else if ((s.a.scope_level > 0 && flags.w_shadow_local) ||
		       (s.a.scope_level == 0 && flags.w_shadow_global)) {
		warn("symbol `" + name + "' shadows other symbol");
		warnl(x.sloc, "(this is the location of previous definition)");
	    }
	}), List::concat(Hashtbl::find_all(c.global_vars, s.name), 
			 Hashtbl::find_all(c.local_vars, s.name)));
    
    if (c.current_scope_level == 0)
	Hashtbl::add(c.global_vars, s.name, s);
    else {
	Hashtbl::add(c.local_vars, s.name, s);
	c.scope_vars = Cons(s, c.scope_vars);
    }

    if (flags.w_shadow_xmodule &&
	((s.a.scope_level > 0 && flags.w_shadow_local) ||
	 (s.a.scope_level == 0 && flags.w_shadow_global))) {
	void wrn(id m)
	{
	    if (m == current_module)
		return;
		
	    try {
		_ x = Module::lookup_var(m, s.name);
		warn("definition of symbol `" + symbol_name(s) + 
		     "' shadows symbol from `" + m + "'");
		warnl(x.sloc, "(defined here)");
	    } with {
	    case Not_found:
		skip;
	    }
	}

	List::iter(wrn, c.open_modules);
    }

    return s;
}

void bind_tydef(bind_ctx c, ty_symbol tydef, ty t)
{
    void add_struct_field(<sf_symbol>field fld)
    {
	make_struct_field(c, tydef, fld);
    }
    
    void add_union_field(<var_symbol>field fld)
    {
	let s = make_var(c, fld.fld_type, fld.fld_symbol) in {
	    set_property(s, Is_union_field);
	    s.a.parent_union = Some(tydef);
	    fld.fld_symbol = Symbol(s);
	}
    }

    void f(ty tt)
    {
	switch tt {
	case Struct(flds):
	    List::iter(add_struct_field, flds);
	    
	case Opt_struct(flds):
	    List::iter(add_struct_field, flds);

	case Union(flds):
	    List::iter(add_union_field, flds);

	case _:
	    skip;
	}
    }

    type_iter(f, t);
}

void push_ty_symbol(bind_ctx c, ty_symbol tysym)
{
    if (tysym.module == "")
	tysym.module = current_module;

    check_if_in_current_module(c, tysym);

    void maybe_redef_error(ty_symbol x)
    {
	switch x.stype {
	case Abstract_type:
	    if (List::length(x.a.tysym_tyvars) !=
		List::length(tysym.a.tysym_tyvars)) {
		error("different number of type arguments " +
		      "then in forward declaration");
		errorl(x.sloc, "here");
	    }
	    x.stype = tysym.stype;

	case _:
	    error("redefinition of type `" + tysym.name + "'");
	    errorl(x.sloc, "first defined here");
	}
    }

    bool check_redef(ty_symbol x) 
    {
	maybe_redef_error(x);
	
	x.a.tysym_tyvars = tysym.a.tysym_tyvars;
	x.stype = tysym.stype;
	tysym = x;

	return true;
    }


    if (!doing_iface)
	try {
	    maybe_redef_error(Module::lookup_tysym(tysym.module, tysym.name));
	} with { case Not_found: skip; }
    
    if (!List::exists(check_redef, Hashtbl::find_all(c.tydefs, tysym.name)))
	Hashtbl::add(c.tydefs, tysym.name, tysym);

    if (flags.w_shadow_xmodule) {
	void wrn(id m)
	{
	    if (m == current_module)
		return;
		
	    try {
		_ s = Module::lookup_tysym(m, tysym.name);
		warn("definition of type `" + symbol_name(tysym) + 
		     "' shadows type from `" + m + "'");
		warnl(s.sloc, "(defined here)");
	    } with {
	    case Not_found:
		skip;
	    }
	}

	List::iter(wrn, c.open_modules);
    }

    bind_tydef(c, tysym, tysym.stype);
}

<'a>symbol lookup_symbol(bind_ctx c, <<'a>symbol>bindable bound_name, 
			 *(<'a>symbol (id)) local_lookup,
			 *(<'a>symbol (id, id)) global_lookup)
{

    switch bound_name {
    case Simple_name(name):
	try {
	    return local_lookup(name);
	} with {
	case Not_found:
	    <<'a>symbol>option ret = None;
	    
	    bool lookup(id mod)
	    {
		try {
		    ret = Some(global_lookup(mod, name));
		    return true;
		} with {
		case Not_found:
		    return false;
		}
	    }
	    
	    if (List::exists(lookup, c.open_modules)) {
		switch ret {
		case Some(s):
		    Hashtbl::replace(c.used_open_modules, s.module, ());
		    return s;
		case None:
		    return ice("Bind::lookup_symbol(): wtf?");
		}
	    } else {
		// sorry Winetou ;)
		raise Not_found;
	    }
	}
	
    case Qualified_name(m, n):
	return global_lookup(m, n);
    
    case Symbol(s):
	return s;
    }
}

ty_symbol lookup_tydef(bind_ctx c, <ty_symbol>bindable name)
{
    ty_symbol r;

    r = lookup_symbol(c, name, partial(Hashtbl::find, c.tydefs),
		      Module::lookup_tysym);
		      
    reference_module(r.module);
    
    return r;
}

var_symbol lookup_var(bind_ctx c, <var_symbol>bindable bname)
{
    var_symbol lookup(id name)
    {
	try {
	    return Hashtbl::find(c.local_vars, name);
	} with {
	case Not_found:
	    return Hashtbl::find(c.global_vars, name);
	}
    }
    
    var_symbol s = lookup_symbol(c, bname, lookup, Module::lookup_var);

    if (has_property(s, Is_union_field)) {
	reference_module(s.module);
	if (has_property(s, Is_exception)) {
	    if (!Hashtbl::mem(global_ctx.used_exns, s.a.extern_name))
		Hashtbl::add(global_ctx.used_exns, s.a.extern_name, ());
	}
    } else
	use_module(s.module);

    set_property(s, Is_used);

    return s;
}

void bind_type(bind_ctx c, ty t)
{
    <ty>list extend_or_cut(int n, <ty>list lst)
    {
	if (n == 0)
	    return Nil;
	    
	switch lst {
	case Nil:
	    return Cons(Abstract_type, extend_or_cut(n - 1, lst));
	    
	case Cons(hd, tl):
	    return Cons(hd, extend_or_cut(n - 1, tl));
	}
    }
    
    void check_parms(nt_record nt)
    {
	let ts = get_binding(nt.tydef) in
	let act_len = List::length(nt.parms) in
	let frm_len = List::length(ts.a.tysym_tyvars) in
	if (act_len != frm_len) {
	    error("number of type arguments of `" + 
		  symbol_name(ts) + "' doesn't match definition");
	    errorl(ts.sloc, "(definition was here)");
	    nt.parms = extend_or_cut(frm_len, nt.parms);
	}
    }
    
    void f(ty tt)
    {
	switch tt {
	case Named_type(nt):
	    try {
		let tydef = lookup_tydef(c, nt.tydef) in {
		    nt.tydef = Symbol(tydef);
		    reference_module(tydef.module);
		}
		check_parms(nt);
	    } with {
	    case Not_found:
		error("unbound type `" + qualified_name(nt.tydef) + "'");
		let (m, n) = fetch_module_name(c, nt.tydef) in
		let artificial_tydef = {name = n, module = m,
					stype = Abstract_type,
					sloc = top_location(),
					a = {tysym_tyvars = Nil}} in
		nt.tydef = Symbol(artificial_tydef);
	    }

	case Func(ft):
	    ft.tyvars = fixup_tyvars(c, ft.tyvars);

	case _:
	    skip;
	}
    }

    type_iter(f, t);
}

var_symbol enter_vardef(bind_ctx c, vardef v)
{
    let s = make_var(c, v.var_type, v.var_symbol) in {
	v.var_symbol = Symbol(s);
	return s;
    }
}

var_symbol bind_vardef(bind_ctx c, vardef v)
{
    bind_type(c, v.var_type);
    return enter_vardef(c, v);
}

bind_ctx make_fn_ctx(bind_ctx c, ty fn_type)
{
    bind_ctx fn_ctx = copy_context(c);
    fn_ctx.formal_parms = Nil;
    fn_ctx.current_fnc_level++;
    Array::set(fn_ctx.closures, fn_ctx.current_fnc_level, Nil);

    void enter_parm(*(ty, <id>option) p)
    {
	switch p {
	case (t, Some(n)):
	    let v = ({var_symbol = Simple_name(n), var_type = t}) in
	    let s = bind_vardef(fn_ctx, v) in {
		set_property(s, Is_lvalue);
		set_property(s, Is_formal_parm);
		set_property(s, Is_initialized);
		fn_ctx.formal_parms = Cons(s, fn_ctx.formal_parms);
	    }
	case (_, None):
	    skip;
	}
    }

    let ft = get_ft_record(fn_type) in {
	bind_type(fn_ctx, fn_type);
	fn_ctx.bound_tyvars = List::concat(ft.tyvars, fn_ctx.bound_tyvars);
	List::iter(enter_parm, ft.parms);
    }

    fn_ctx.formal_parms = List::rev(fn_ctx.formal_parms);

    return fn_ctx;
}

bind_ctx make_ty_ctx(bind_ctx c, ty_symbol ts)
{
    bind_ctx ty_ctx = copy_context(c);

    ty_ctx.current_scope_level = c.current_scope_level;
    ty_ctx.bound_tyvars = List::concat(ts.a.tysym_tyvars, c.bound_tyvars);

    return ty_ctx;
}

void warn_unused(bind_ctx c)
{
    void warn_one(var_symbol s)
    {
	if (has_property(s, Is_used)) 
	    return;

	if (has_property(s, Is_formal_parm)) {
	    if (flags.w_unused_parm)
		warnl(s.sloc, "unused function parameter `" + s.name + "'");
	} else if (has_property(s, Is_in_pattern)) {
	    if (flags.w_unused_patvar)
		warnl(s.sloc, "unused pattern variable `" + s.name + "'");
	} else {
	    if (flags.w_unused_variable)
		warnl(s.sloc, "unused local variable `" + s.name + "'");
	}
    }

    List::iter(warn_one, c.scope_vars);
}

void bind_lambda(bind_ctx c, lambda_record l)
{
    bind_ctx fn_ctx = make_fn_ctx(c, l.lambda_type);
    
    l.lambda_parms = fn_ctx.formal_parms;

    if (fn_ctx.current_fnc_level + 2 >= Array::length(c.closures))
	Array::resize(c.closures, fn_ctx.current_fnc_level * 3, Nil);
	
    bind_stmt(fn_ctx, l.lambda_body);
	
    l.lambda_closure_vars = 
	    Array::get(fn_ctx.closures, fn_ctx.current_fnc_level);
    l.lambda_needs_closure = fn_ctx.needs_closure;

    // sanity checks
    switch (l.lambda_needs_closure, l.lambda_closure_vars) {
    case (false, Cons(_, _)):
	ice("Bind::bind_lambda(): needs no closure, but has closure vars");

    case (true, Nil):
	if (fn_ctx.current_fnc_level == 1)
	    ice("Bind::bind_lambda(): needs no closure, but need one");

    case _:
	skip;
    }

    warn_unused(fn_ctx);
}

<sf_symbol>list lookup_struct_field(bind_ctx c, <sf_symbol>bindable name)
{
    <sf_symbol>list find_all(<sf_symbol>list acc, 
			        <id>list modules, id n)
    {
	switch modules {
	case Cons(m, tl):
	    let th = Module::lookup_struct_field(m, n) in {
		if (List::length(th) > 0)
		    Hashtbl::replace(c.used_open_modules, m, ());
		return find_all(List::concat(th, acc), tl, n);
	    }
	case Nil:
	    return acc;
	}
    }
    
    switch name {
    case Symbol(s):
	return List::cons1(s);
	
    case Qualified_name(m, n):
	reference_module(m);
	return Module::lookup_struct_field(m, n);
    
    case Simple_name(n):
	let here = Hashtbl::find_all(c.struct_fields, n) in
	let there = find_all(Nil, c.open_modules, n) in
	return List::concat(here, there);
    }
}

<ty_symbol>list possible_structs(bind_ctx c, <<sf_symbol>bindable>list syms)
{
    <ty_symbol>list possible_tydefs = Nil;

    <ty_symbol>list lookup_tydefs(<sf_symbol>bindable name)
    {
	return List::map(fun ty_symbol (sf_symbol s) (s.a.sf_tydef), 
			 lookup_struct_field(c, name));
    }
    
    bool check_if_possible(ty_symbol t)
    {
	return List::length(possible_tydefs) == 0 ||
	       List::exists(fun bool (ty_symbol p) 
				(p.name == t.name && p.module == t.module),
			    possible_tydefs);
    }
    
    void bind_one(<sf_symbol>bindable s)
    {
	let new_tydefs = List::filter(check_if_possible, lookup_tydefs(s)) in
	if (List::length(new_tydefs) == 0) {
	    string names = 
	      String::concat(" nor ", List::map(symbol_name, possible_tydefs));
	    if (names == "")
		names = "any structure";
			     
	    error("field `" + qualified_name(s) + "' is not part of " + names);
	} else {
	    possible_tydefs = new_tydefs;
	}
    }

    List::iter(bind_one, syms);

    return possible_tydefs;
}

expr make_expr(raw_expr r)
{
    return {loc = top_location(), a = {raw = r, expr_type = None}};
}

void bind_ref(bind_ctx c, nl_expr ex, ref_record rr)
{
    try {
	var_symbol s = lookup_var(c, rr.ref_symbol);

	if (has_property(s, Is_union_field)) {
	    // this is sth like Nil, we need to change
	    // this expression node to reflect this fact.
	    ex.raw = Union_expr({nx_symbol = Symbol(s), 
				nx_expr = make_expr(Void_const)});
	    return;
	}

	rr.ref_symbol = Symbol(s);
	
	if (s.a.fnc_level > 0)
	    rr.ref_level = c.current_fnc_level - s.a.fnc_level;
	else
	    rr.ref_level = 0;
	    
	if (rr.ref_level > 0) {
	    set_property(s, Is_in_closure);
	    
	    let lst = Array::get(c.closures, s.a.fnc_level);
	    if (!List::exists(fun (x) (Hacks::same_ptrs(s, x)), lst))
		Array::set(c.closures, s.a.fnc_level, Cons(s, lst));
	    
	    bind_ctx p = c;
	    while (p.current_fnc_level >= s.a.fnc_level) {
		if (p.current_fnc_level < c.current_fnc_level)
		    p.needs_closure = true;
		switch p.parent {
		case Some(pp):
		    p = pp;
		case None:
		    ice("Bind::bind_ref(): end of chain");
		}
	    }
	}
    } with {
    case Not_found:
	error("unbound variable `" + qualified_name(rr.ref_symbol) + "'");
	raise Choke;
    }
}

var_symbol bind_union_tag(bind_ctx c, <var_symbol>bindable sym)
{
    try {
	let s = lookup_var(c, sym) in 
	if (!has_property(s, Is_union_field)) {
	    error("`" + qualified_name(sym) +
		  "' is not part of any union");
	    raise Choke;
	} else {
	    return s;
	}
    } with {
    case Not_found:
	error("unbound union field `" + qualified_name(sym) + "'");
	raise Choke;
    }
}

void bind_call(bind_ctx c, nl_expr ex, expr fnc, <expr>list args)
{
    bind_expr(c, fnc);
    List::iter(partial(bind_expr, c), args);
    
    switch fnc.a.raw {
    case Union_expr(u):
	// We really have an Union_expr with arguments here.
	// There used to be Ref(...) as fnc here, but bind_ref() put Union_expr
	// with Void_const argument, now we just need to fix up the arguments.
	switch args {
	case Nil:
	    skip;	// nothing to do
	case Cons(a, Nil):
	    u.nx_expr = a;
	case _:
	    u.nx_expr = make_expr(Tuple_expr(args));
	}

	// anyway we need to store Union_expr one level upper then it is
	ex.raw = fnc.a.raw;
	
    case _:
	skip;
    }
}

void bind_raw_expr(bind_ctx c, nl_expr ex)
{
    switch ex.raw {
    case Assign(e1, e2):
    case Simple_binop(_, e1, e2):
	bind_expr(c, e1);
	bind_expr(c, e2);
	
    case Simple_unop(_, e):
	bind_expr(c, e);
    
    case Ref(rr):
	bind_ref(c, ex, rr);

    case Call(e, el):
	bind_call(c, ex, e, el);

    case Tuple_expr(el):
	List::iter(partial(bind_expr, c), el);

    case Union_expr(u):
	bind_expr(c, u.nx_expr);
	u.nx_symbol = Symbol(bind_union_tag(c, u.nx_symbol));
    
    case Lambda(l):
	bind_lambda(c, l);

    case Struct_expr(ss, sfb):
	let syms = List::map(fun (s) (s.sx_symbol), ss) in
	sfb.sfb_tydefs = possible_structs(c, syms);
	List::iter(fun (s) (bind_expr(c, s.sx_expr)), ss);
	switch sfb.sfb_tydefs {
	case [_]:
	case Nil:
	    skip;
	case Cons(_, _) as l:
	    let names = String::concat(", ", List::map(symbol_name, l));
	    error("more then one type possible for expression: " + names);
	}

    // special case, leave sfb with Nil
    case Struct_copy(e, Nil, _):
	bind_expr(c, e);

    case Struct_copy(e, ss, sfb):
	let syms = List::map(fun (s) (s.sx_symbol), ss) in
	sfb.sfb_tydefs = possible_structs(c, syms);
	bind_expr(c, e);
	List::iter(fun (s) (bind_expr(c, s.sx_expr)), ss);

    case Dot_expr(de, sfb):
	sfb.sfb_tydefs = possible_structs(c, [de.sx_symbol]);
	bind_expr(c, de.sx_expr);
    
    // contants arn't really interesting in this context ;)
    case Bool_const(_):
    case Int_const(_):
    case Float_const(_):
    case String_const(_): 
    case Void_const:
    case Null_const:
    case Assign_target:
	skip;
    }

}

void bind_expr(bind_ctx c, expr e)
{
    locate(partial(bind_raw_expr, c), e);
}

bool is_capitalized(string s)
{
    return int_of_char(Unitype::to_uppercase(String::get(s, 0))) == 
	   int_of_char(String::get(s, 0));
}

void bind_pattern_int(bind_ctx c, pattern p)
{
    switch p.a {
    case Void_pattern:
	skip;
	
    case Var_pattern(v):
	let s = bind_vardef(c, v) in {
	    if (flags.w_capitalization && 
		is_capitalized(simple_name(v.var_symbol)))
		warn("capitalized pattern variable `" +
		     simple_name(v.var_symbol) + "'");
	    set_property(s, Is_in_pattern);
	    c.pattern_vars = Cons(s, c.pattern_vars);
	}

    case As_pattern(pat, v):
	let s = bind_vardef(c, v) in {
	    if (flags.w_capitalization && 
		is_capitalized(simple_name(v.var_symbol)))
		warn("capitalized pattern variable `" +
		     simple_name(v.var_symbol) + "'");
	    set_property(s, Is_in_pattern);
	    c.pattern_vars = Cons(s, c.pattern_vars);
	}

	bind_pattern_int(c, pat);

    case Expr_pattern(e):
	/* 
	 * Expr_pattern(...) is generated for `Nil', `Nil', 
	 * `x' and `y' where Nil and Nil are union members, x 
	 * is some bound variable and y is free variable. We fixup 
	 * it here. Note that type of Var_pattern(..) vardef is set
	 * to Abstract_type. This is because we cannot tell the
	 * actual type here (at least not in all cases).
	 */
	switch e.a.raw {
	case Ref(rr):
	    try {
		let s = lookup_var(c, rr.ref_symbol) in
		if (has_property(s, Is_union_field)) {
		    let np = {np_symbol = Symbol(s),
			      np_pattern = {loc = top_location(), 
					    a = Void_pattern()}} in
		    p.a = Union_pattern(np);
		    return;
		}
	    } with {
	    case Not_found:
		skip;
	    }
	    
	    let xt = genxt("_patvar") in
	    p.a = Var_pattern({var_symbol = rr.ref_symbol, 
			       var_type = X_tyvar(xt)});
	    bind_pattern_int(c, p);
	    
	case _:
	    bind_expr(c, e);
	}

    // Fixup 0 and 1-long tuple patterns
    case Tuple_pattern(Nil):
	p.a = Void_pattern;

    case Tuple_pattern(Cons(x, Nil)):
	p.loc = x.loc;
	p.a = x.a;
	bind_pattern_int(c, p);
	
    case Tuple_pattern(lst):
	List::iter(partial(bind_pattern, c), lst);
    
    case Union_pattern(np):
	np.np_symbol = Symbol(bind_union_tag(c, np.np_symbol));
	bind_pattern_int(c, np.np_pattern);

    case Struct_pattern(sp):
	let syms = List::map(fun (s) (s.sp_symbol), sp.sp_entries) in
	sp.sp_bindings.sfb_tydefs = possible_structs(c, syms);
	List::iter(fun (s) (bind_pattern(c, s.sp_pattern)), sp.sp_entries);
    }
}

void bind_pattern(bind_ctx c, pattern p)
{
    push_location(p.loc);
    try {
	bind_pattern_int(c, p);
    } finally {
	pop_location();
    }
}

struct pattern_var_info {
    var_symbol primary_symbol;
    int occurred_cnt;
}

void bind_case_clause(bind_ctx c, case_clause cc)
{
    <pattern_var_info>Hashtbl::t vars = Hashtbl::create(13);
    bool first = true;
    int ncases = List::length(cc.cases);

    <var_symbol>list bind_one_case_1(*(pattern, expr) oc)
    {
	void add_to_hash(var_symbol s)
	{
	    if (first) {
		Hashtbl::add(vars, s.name,
			     {primary_symbol = s, occurred_cnt = 1});
	    } else {
		try {
		    let i = Hashtbl::find(vars, s.name) in
		    i.occurred_cnt++;
		} with { case Not_found: skip; }
	    }
	}
	
	let (pat, _) = oc in
	let ctx = copy_context(c) in {
	    if (List::length(ctx.pattern_vars) != 0)
		ice("Bind::bind_case_clause(): pattern_vars not clear");
	    bind_pattern(ctx, pat);
	    List::iter(add_to_hash, ctx.pattern_vars);
	    first = false;
	    return ctx.pattern_vars;
	}
    }

    var_symbol maybe_change(var_symbol s)
    {
	try {
	    let i = Hashtbl::find(vars, s.name) in
	    if (i.occurred_cnt == ncases)
		return i.primary_symbol;
	} with { case Not_found: skip; }

	set_property(s, Is_guard_local);
	return s;
    }
    
    void bind_one_case_2(*(pattern, expr) oc, <var_symbol>list patvars)
    {
	void change_pattern(pattern p)
	{
	    switch p.a {
		case Var_pattern(v):
		    let s = get_binding(v.var_symbol) in {
			v.var_symbol = Symbol(maybe_change(s));
			v.var_type = s.stype;
		    }

		case As_pattern(pat, v):
		    let s = get_binding(v.var_symbol) in {
			v.var_symbol = Symbol(maybe_change(s));
			v.var_type = s.stype;
		    }
		    change_pattern(pat);
		
		case Void_pattern:	skip;
		case Expr_pattern(_):    skip;
		
		case Tuple_pattern(lst):
		    List::iter(change_pattern, lst);
		
		case Struct_pattern(sp):
		    List::iter(fun (s) (change_pattern(s.sp_pattern)), 
			       sp.sp_entries);
		
		case Union_pattern(np):
		    change_pattern(np.np_pattern);
	    }
	}

	let (pat, cond) = oc in
	let ctx = copy_context(c) in {
	    void push_symbol(var_symbol s)
	    {
		s = maybe_change(s);
		Hashtbl::add(ctx.local_vars, s.name, s);
	    }

	    List::iter(push_symbol, patvars);
	    bind_expr(ctx, cond);
	    
	    let gloc_p = fun (s) (has_property(s, Is_guard_local)) in {
		ctx.scope_vars = List::filter(gloc_p, ctx.scope_vars);
		warn_unused(ctx);
	    }

	    change_pattern(pat);
	}
    }

    let patvars_lists = List::map(bind_one_case_1, cc.cases) in
    List::iter2(bind_one_case_2, cc.cases, patvars_lists);
    
    let ctx = copy_context(c) in {
	void maybe_push(string, pattern_var_info i)
	{
	    if (i.occurred_cnt == ncases) {
		let s = i.primary_symbol in {
		    Hashtbl::add(ctx.local_vars, s.name, s);
		    cc.patvars = Cons(s, cc.patvars);
		    ctx.scope_vars = Cons(s, ctx.scope_vars);
		}
	    }
	}
	
	Hashtbl::iter(vars, maybe_push);

	switch cc.body.a {
	case Block(Nil):
	    ice("Bind::bind_case_clause(): empty case body");
	case _:
	    bind_stmt(ctx, cc.body);
	    warn_unused(ctx);
	}
    }
}

void open_module(bind_ctx c, id mod)
{
    if (!List::exists(fun bool (id x) (x == mod), c.open_modules))
	try {
	    // error-check
	    Module::open_module(mod);
	    c.open_modules = Cons(mod, c.open_modules);
	    reference_module(mod);
	} with {
	case Choke:
	    // just don't open it.
	    skip;
	}
}

void bind_loop(bind_ctx c, loop_record l)
{
    string name;

    switch l.ll_symbol {
    case Symbol(_):
	ice("Bind::bind_loop(): bound");
	return;
	
    case Qualified_name(_, _):
	ice("Bind::bind_loop(): qn");
	return;
    
    case Simple_name(n):
	name = n;
    }
    
    if (name != "" && flags.w_shadow_label && Hashtbl::mem(c.loop_ids, name)) {
	warn("loop label `" + name + "' shadows another loop label");
	warnl((Hashtbl::find(c.loop_ids, name)).sloc, "(first defined here)");
    }

    let loop_ctx = copy_context(c) in {
	loop_ctx.current_loop_level = c.current_loop_level + 1;
	
	_ sym = {module = "", 
		 name = name, 
		 stype = Abstract_type,
		 sloc = top_location(), 
		 a = {loop_level = loop_ctx.current_loop_level, is_used = false}};
		 
	Hashtbl::add(loop_ctx.loop_ids, sym.name, sym);
	l.ll_symbol = Symbol(sym);
	loop_ctx.top_loop_id = Some(sym);
    
	bind_stmt(copy_context(loop_ctx), l.ll_body);
	bind_stmt(copy_context(loop_ctx), l.ll_inc);

	if (flags.w_unused_label && sym.name != "" && !sym.a.is_used)
	    warn("unused loop label `" + sym.name + "'");
    }
}

void bind_break(bind_ctx c, break_record b)
{
    string name;

    switch b.break_symbol {
    case Symbol(_):
	ice("Bind::bind_break(): bound");
	return;
	
    case Qualified_name(_, _):
	ice("Bind::bind_break(): qn");
	return;
    
    case Simple_name(n):
	name = n;
    }

    switch c.top_loop_id {
    case Some(top):
	if (name == "") {
	    b.break_symbol = Symbol(top);
	    if (top.name != "" && flags.w_labeled) {
		warn("break without a label in labeled (with `" + top.name +
		     "') loop");
	    }
	} else {
	    try {
		let s = Hashtbl::find(c.loop_ids, name) in {
		    b.break_symbol = Symbol(s);
		    s.a.is_used = true;
		}
	    } with {
	    case Not_found:
		error("unbound loop label `" + name + "'");
		b.break_symbol = Symbol(top);
	    }
	}
    case None:
	error("break outside loop");
    }
}

void bind_let_pattern(bind_ctx c, pattern pat)
{
    if (List::length(c.pattern_vars) != 0)
	ice("Bind::bind_let_pattern(): pattern_vars not clear");
	
    bind_pattern(c, pat);
    
    List::iter(fun void (var_symbol s) {set_property(s, Is_let_variable);},
	       c.pattern_vars);

    c.pattern_vars = Nil;
}

void bind_nl_stmt(bind_ctx c, nl_stmt r)
{
    void bind_in_scope(stmt s)
    {
	let sc = copy_context(c) in {
	    bind_stmt(sc, s);
	    warn_unused(sc);
	}
    }
    
    switch r {
    case Skip:
	skip;
    
    case Expr(e):
	bind_expr(c, e);
    
    case Return(e):
	bind_expr(c, e);

    case Vardef(v):
	let s = bind_vardef(c, v) in
	set_property(s, Is_lvalue);
	
    case Block(lst):
	let block_ctx = copy_context(c) in {
	    List::iter(partial(bind_stmt, block_ctx), lst);
	    warn_unused(block_ctx);
	}

    case Pair(s1, s2):
	bind_stmt(c, s1);
	bind_stmt(c, s2);
    
    case Switch(e, clauses):
	bind_expr(c, e);
	List::iter(partial(bind_case_clause, c), clauses);
	
    case Let(assigns, s):
	let ctx = copy_context(c) in {
	    List::iter(partial(bind_expr, ctx), List::map(snd, assigns));
	    List::iter(partial(bind_let_pattern, ctx), List::map(fst, assigns));
	    bind_stmt(ctx, s);
	    warn_unused(ctx);
	}
    
    case Flying_let(assigns):
	List::iter(partial(bind_expr, c), List::map(snd, assigns));
	List::iter(partial(bind_let_pattern, c), List::map(fst, assigns));
    
    case If(e, s1, s2):
	bind_expr(c, e);
	bind_in_scope(s1);
	bind_in_scope(s2);

    case Loop(l):
	bind_loop(c, l);
    
    case Break(b):
	bind_break(c, b);
	
    case Break_if_not(e):
	bind_expr(c, e);
    
    case Open(m):
	open_module(c, m);
    
    case Try(tr):
	bind_in_scope(tr.try_body);
	List::iter(partial(bind_case_clause, c), tr.try_with);
	bind_in_scope(tr.try_finally);
    
    case Raise(e):
	bind_expr(c, e);
    }
}

void bind_stmt(bind_ctx c, stmt s)
{
    try {
	locate(partial(bind_nl_stmt, c), s);
    } with {
    case Choke:
	skip;
    }
}

void iface_check(string m)
{
    if (doing_iface)
	error(m + " not allowed in interface");
}

// enter top-level symbol into binding context
void enter_nl_decl(nl_decl d)
{
    check_init();
    switch d {
    case Func_decl(fn):
	iface_check("function definition");
	let s = make_var(global_ctx, fn.fn_lambda.lambda_type, fn.fn_symbol) in {
	    fn.fn_symbol = Symbol(s);
	    if (fn.fn_export_as != "")
		s.a.extern_name = fn.fn_export_as;
	}
	
    case Global_decl(v):
	let s = enter_vardef(global_ctx, v) in {
	    set_property(s, Is_lvalue);
	    set_property(s, Is_global_var);
	}
    
    case Extern_c(xn, v):
	let s = enter_vardef(global_ctx, v) in {
	    if (xn == "")
		s.a.extern_name = s.name;
	    else
		s.a.extern_name = xn;
	    set_property(s, Is_extern_c);
	    switch s.stype {
	    case Func(_):
		skip;
	    case _:
		set_property(s, Is_lvalue);
	    }
	}
    
    case Func_proto(v):
	enter_vardef(global_ctx, v);

    case Top_level_open(_):
	skip;
    
    case Tydef(tysym):
	push_ty_symbol(global_ctx, tysym);
    
    case Exception_decl(v):
	void poly_check(ty t)
	{
	    switch t {
	    case Tyvar(_):
		error("exception types cannot be polymorphic");
	    case _:
		skip;
	    }
	}

	type_iter(poly_check, v.var_type);
	let s = enter_vardef(global_ctx, v) in {
	    set_property(s, Is_exception);
	    set_property(s, Is_union_field);
	}
    
    case Section_decl(i, _):
	iface_check("section definition");
	switch i {
	case "init":	skip;
	case "fini":	skip;
	case _:
	    error("unknown section `" + i + "'");
	}
    }
}

void enter_decl(decl d)
{
    locate(enter_nl_decl, d);
}

void bind_sym_type(bind_ctx c, <<'a>symbol>bindable sym)
{
    switch sym {
    case Symbol(s):
	bind_type(c, s.stype);
    
    case _:
	ice("bind_sym_type(): not bound");
    }
}

void bind_nl_decl(nl_decl d)
{
    check_init();
    switch d {
    case Top_level_open(m):
	open_module(global_ctx, m);
    
    case Func_decl(fn):
	bind_lambda(global_ctx, fn.fn_lambda);
    
    case Section_decl(_, l):
	bind_lambda(global_ctx, l);
    
    // here just need to bind type
    case Global_decl(v):
	bind_sym_type(global_ctx, v.var_symbol);
    
    case Extern_c(_, v):
	bind_sym_type(global_ctx, v.var_symbol);
    
    case Func_proto(v):
	bind_sym_type(global_ctx, v.var_symbol);

    case Tydef(tysym):
	bind_type(make_ty_ctx(global_ctx, tysym), tysym.stype);
    
    case Exception_decl(v):
	bind_sym_type(global_ctx, v.var_symbol);
    }
}

void bind_decl(decl d)
{
    locate(bind_nl_decl, d);
}

void finish_bind()
{
    if (flags.w_unused_open) {
	_ c = global_ctx;
	_ unused = List::filter(fun (n) (!Hashtbl::mem(c.used_open_modules, n)), 
				c.open_modules);
	if (List::length(unused) != 0)
	    warn("open declarations for following modules are unused: `" +
		 String::concat("', `", unused) + "'");
    }
}

<id>list get_modules(<void>Hashtbl::t h)
{
    <id>list ret = Nil;
    
    void add(string name, void)
    {
	if (name != current_module && name != "")
	    ret = Cons(name, ret);
    }
    
    Hashtbl::iter(h, add);

    return ret;
}

<id>list used_modules()
{
    check_init();
    return get_modules(global_ctx.used_modules);
}

<id>list referenced_modules()
{
    check_init();
    return get_modules(global_ctx.referenced_modules);
}

<id>list used_exns()
{
    check_init();
    return get_modules(global_ctx.used_exns);
}

void save_module(string filename)
{
    check_init();

    Module::module m = 
	   ({Module::mod_name = current_module,
	     Module::mod_under_construction = false,
	     Module::mod_vars = global_ctx.global_vars,
	     Module::mod_struct_fields = global_ctx.struct_fields,
	     Module::mod_tydefs = global_ctx.tydefs,
	     Module::mod_depends_on = referenced_modules()
	    });
    Module::save_module(m, filename);
}

// vim:sts=4:sw=4
